# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import time
import shutil
import platform
import tempfile
import unittest
import multiprocessing

from libcloud.utils.files import exhaust_iterator
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Object, Container
from libcloud.storage.types import (
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)

try:
    import fasteners  # noqa

    from libcloud.storage.drivers.local import LockLocalStorage, LocalStorageDriver
except ImportError:
    print("fasteners library is not available, skipping local_storage tests...")
    LocalStorageDriver = None


class PickleableAcquireLockInSubprocess:
    def __call__(self, pid, success):
        # For first process acquire should succeed and for the second it should fail
        lock = LockLocalStorage("/tmp/c", timeout=0.5)

        if pid == 1:
            with lock:
                # We use longer sleep when running tests in parallel to avoid
                # failures related to slower process spawn
                time.sleep(2.5)

            success.value = 1
        elif pid == 2:
            expected_msg = "Failed to acquire IPC lock"
            try:
                lock.__enter__()
            except LibcloudError as e:
                assert expected_msg in str(e)
                success.value = 1
        else:
            raise ValueError("Invalid pid")


class LocalTests(unittest.TestCase):
    driver_type = LocalStorageDriver

    @classmethod
    def create_driver(self):
        self.key = tempfile.mkdtemp()
        return self.driver_type(self.key, None)

    def setUp(self):
        self.driver = self.create_driver()

    def tearDown(self):
        shutil.rmtree(self.key)
        self.key = None

    def make_tmp_file(self, content=None):
        if not content:
            content = b"blah" * 1024
        _, tmppath = tempfile.mkstemp()
        with open(tmppath, "wb") as fp:
            fp.write(content)
        return tmppath

    def remove_tmp_file(self, tmppath):
        try:
            os.unlink(tmppath)
        except Exception as e:
            msg = str(e)
            if "being used by another process" in msg and platform.system().lower() == "windows":
                return
            raise e

    @unittest.skipIf(platform.system().lower() == "windows", "Unsupported on Windows")
    def test_lock_local_storage(self):
        # 1. Acquire succeeds
        lock = LockLocalStorage("/tmp/a")
        with lock:
            self.assertTrue(True)

        # 2. Acquire fails because lock is already acquired
        lock = LockLocalStorage("/tmp/b", timeout=0.5)
        with lock:
            expected_msg = "Failed to acquire thread lock"
            self.assertRaisesRegex(LibcloudError, expected_msg, lock.__enter__)

        success_1 = multiprocessing.Value("i", 0)
        success_2 = multiprocessing.Value("i", 0)

        p1 = multiprocessing.Process(
            target=PickleableAcquireLockInSubprocess(),
            args=(
                1,
                success_1,
            ),
        )
        p1.start()

        time.sleep(0.2)

        p2 = multiprocessing.Process(
            target=PickleableAcquireLockInSubprocess(),
            args=(
                2,
                success_2,
            ),
        )
        p2.start()

        p1.join()
        p2.join()

        self.assertEqual(bool(success_1.value), True, "Check didn't pass")
        self.assertEqual(bool(success_2.value), True, "Second check didn't pass")

    def test_list_containers_empty(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

    def test_containers_success(self):
        self.driver.create_container("test1")
        self.driver.create_container("test2")
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 2)

        container = containers[1]

        self.assertTrue("creation_time" in container.extra)
        self.assertTrue("modify_time" in container.extra)
        self.assertTrue("access_time" in container.extra)

        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

        objects = container.list_objects()
        self.assertEqual(len(objects), 0)

        for container in containers:
            self.driver.delete_container(container)

    def test_objects_success(self):
        tmppath = self.make_tmp_file()

        container = self.driver.create_container("test3")
        obj1 = container.upload_object(tmppath, "object1")
        obj2 = container.upload_object(tmppath, "path/object2")
        obj3 = container.upload_object(tmppath, "path/to/object3")
        obj4 = container.upload_object(tmppath, "path/to/object4.ext")

        with open(tmppath, "rb") as tmpfile:
            obj5 = container.upload_object_via_stream(tmpfile, "object5")

        obj6 = container.upload_object(tmppath, "foo5")
        obj7 = container.upload_object(tmppath, "foo6")
        obj8 = container.upload_object(tmppath, "Afoo7")

        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 8)

        # Prefix filtering
        prefix = os.path.join("path", "invalid")
        objects = self.driver.list_container_objects(container=container, prefix=prefix)
        self.assertEqual(len(objects), 0)

        prefix = os.path.join("path", "to")
        objects = self.driver.list_container_objects(container=container, prefix=prefix)
        self.assertEqual(len(objects), 2)
        self.assertEqual(objects[0].name, "path/to/object3")
        self.assertEqual(objects[1].name, "path/to/object4.ext")

        prefix = "foo"
        objects = self.driver.list_container_objects(container=container, prefix=prefix)
        self.assertEqual(len(objects), 2)
        self.assertEqual(objects[0].name, "foo5")
        self.assertEqual(objects[1].name, "foo6")

        prefix = "foo5"
        objects = self.driver.list_container_objects(container=container, prefix=prefix)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0].name, "foo5")

        prefix = "foo6"
        objects = self.driver.list_container_objects(container=container, prefix=prefix)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0].name, "foo6")

        for obj in objects:
            self.assertNotEqual(obj.hash, None)
            self.assertEqual(obj.size, 4096)
            self.assertEqual(obj.container.name, "test3")
            self.assertTrue("creation_time" in obj.extra)
            self.assertTrue("modify_time" in obj.extra)
            self.assertTrue("access_time" in obj.extra)

        obj1.delete()
        obj2.delete()

        objects = container.list_objects()
        self.assertEqual(len(objects), 6)

        container.delete_object(obj3)
        container.delete_object(obj4)
        container.delete_object(obj5)
        container.delete_object(obj6)
        container.delete_object(obj7)
        container.delete_object(obj8)

        objects = container.list_objects()
        self.assertEqual(len(objects), 0)

        container.delete()
        self.remove_tmp_file(tmppath)

    def test_get_container_doesnt_exist(self):
        try:
            self.driver.get_container(container_name="container1")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_container_success(self):
        self.driver.create_container("test4")
        container = self.driver.get_container(container_name="test4")
        self.assertTrue(container.name, "test4")
        container.delete()

    def test_get_object_container_doesnt_exist(self):
        try:
            self.driver.get_object(container_name="test-inexistent", object_name="test")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_object_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container("test5")
        container.upload_object(tmppath, "test")

        obj = self.driver.get_object(container_name="test5", object_name="test")

        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.container.name, "test5")
        self.assertEqual(obj.size, 4096)
        self.assertNotEqual(obj.hash, None)
        self.assertTrue("creation_time" in obj.extra)
        self.assertTrue("modify_time" in obj.extra)
        self.assertTrue("access_time" in obj.extra)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_create_container_invalid_name(self):
        try:
            self.driver.create_container(container_name="new/container")
        except InvalidContainerNameError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_container_already_exists(self):
        container = self.driver.create_container(container_name="new-container")
        try:
            self.driver.create_container(container_name="new-container")
        except ContainerAlreadyExistsError:
            pass
        else:
            self.fail("Exception was not thrown")

        # success
        self.driver.delete_container(container)

    def test_create_container_success(self):
        name = "new_container"
        container = self.driver.create_container(container_name=name)
        self.assertEqual(container.name, name)
        self.driver.delete_container(container)

    def test_delete_container_doesnt_exist(self):
        container = Container(name="new_container", extra=None, driver=self.driver)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_container_not_empty(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        try:
            self.driver.delete_container(container=container)
        except ContainerIsNotEmptyError:
            pass
        else:
            self.fail("Exception was not thrown")

        # success
        obj.delete()
        self.remove_tmp_file(tmppath)
        self.assertTrue(self.driver.delete_container(container=container))

    def test_delete_container_not_found(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Container does not exist but an exception was not" + "thrown")

    def test_delete_container_success(self):
        container = self.driver.create_container("test7")
        self.assertTrue(self.driver.delete_container(container=container))

    def test_download_object_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        destination_path = tmppath + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )

        self.assertTrue(result)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_and_overwrite(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        destination_path = tmppath + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )

        self.assertTrue(result)

        try:
            self.driver.download_object(
                obj=obj,
                destination_path=destination_path,
                overwrite_existing=False,
                delete_on_failure=True,
            )
        except LibcloudError:
            pass
        else:
            self.fail("Exception was not thrown")

        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=True,
            delete_on_failure=True,
        )

        self.assertTrue(result)

        # success
        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_as_stream_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=1024)

        self.assertTrue(hasattr(stream, "__iter__"))

        data = b"".join(stream)
        self.assertTrue(len(data), 4096)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_download_object_range_success(self):
        content = b"0123456789123456789"
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        destination_path = tmppath + ".temp"

        # 1. Only start_bytes provided
        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=4,
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path, "rb") as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b"456789123456789")
        self.assertEqual(written_content, content[4:])

        # 2. start_bytes and end_bytes is provided
        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=4,
            end_bytes=6,
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path, "rb") as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b"45")
        self.assertEqual(written_content, content[4:6])

        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=0,
            end_bytes=1,
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path, "rb") as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b"0")
        self.assertEqual(written_content, content[0:1])

        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=0,
            end_bytes=2,
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path, "rb") as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b"01")
        self.assertEqual(written_content, content[0:2])

        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=0,
            end_bytes=len(content),
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path, "rb") as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b"0123456789123456789")
        self.assertEqual(written_content, content[0 : len(content)])

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_range_as_stream_success(self):
        content = b"0123456789123456789"
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        # 1. Only start_bytes provided
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=4, chunk_size=1024
        )
        written_content = b"".join(stream)

        self.assertEqual(written_content, b"456789123456789")
        self.assertEqual(written_content, content[4:])

        # 2. start_bytes and end_bytes is provided
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=4, end_bytes=7, chunk_size=1024
        )
        written_content = b"".join(stream)

        self.assertEqual(written_content, b"456")
        self.assertEqual(written_content, content[4:7])

        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=0, end_bytes=1, chunk_size=1024
        )
        written_content = b"".join(stream)

        self.assertEqual(written_content, b"0")
        self.assertEqual(written_content, content[0:1])

        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=1, end_bytes=3, chunk_size=1024
        )
        written_content = b"".join(stream)

        self.assertEqual(written_content, b"12")
        self.assertEqual(written_content, content[1:3])

        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=0, end_bytes=len(content), chunk_size=1024
        )
        written_content = b"".join(stream)

        self.assertEqual(written_content, b"0123456789123456789")
        self.assertEqual(written_content, content[0 : len(content)])

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_download_object_range_invalid_values(self):
        obj = Object("a", 500, "", {}, {}, None, None)
        tmppath = self.make_tmp_file(content="")

        expected_msg = "start_bytes must be greater than 0"
        self.assertRaisesRegex(
            ValueError,
            expected_msg,
            self.driver.download_object_range,
            obj=obj,
            destination_path=tmppath,
            start_bytes=-1,
        )

        expected_msg = "start_bytes must be smaller than end_bytes"
        self.assertRaisesRegex(
            ValueError,
            expected_msg,
            self.driver.download_object_range,
            obj=obj,
            destination_path=tmppath,
            start_bytes=5,
            end_bytes=4,
        )

        expected_msg = "start_bytes and end_bytes can't be the same"
        self.assertRaisesRegex(
            ValueError,
            expected_msg,
            self.driver.download_object_range,
            obj=obj,
            destination_path=tmppath,
            start_bytes=5,
            end_bytes=5,
        )

    def test_download_object_range_as_stream_invalid_values(self):
        content = b"0123456789123456789"
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container("test6")
        obj = container.upload_object(tmppath, "test")

        expected_msg = "start_bytes must be greater than 0"
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=-1, end_bytes=None, chunk_size=1024
        )
        self.assertRaisesRegex(ValueError, expected_msg, exhaust_iterator, stream)

        expected_msg = "start_bytes must be smaller than end_bytes"
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=5, end_bytes=4, chunk_size=1024
        )
        self.assertRaisesRegex(ValueError, expected_msg, exhaust_iterator, stream)

        expected_msg = "end_bytes is larger than file size"
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=5, end_bytes=len(content) + 1, chunk_size=1024
        )

        expected_msg = "start_bytes and end_bytes can't be the same"
        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=5, end_bytes=5, chunk_size=1024
        )

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)


if not LocalStorageDriver:

    class LocalTests(unittest.TestCase):  # NOQA
        pass


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import CONTAINER_PARAMS_RANCHER
from libcloud.container.base import ContainerImage
from libcloud.test.file_fixtures import ContainerFileFixtures
from libcloud.container.drivers.rancher import RancherContainerDriver

# --------------------------------------------------------------------------- #
# Mock Classes


class RancherMockHttp(MockHttp):
    fixtures = ContainerFileFixtures("rancher")

    def _v1_environments(self, method, url, body, headers):
        if method == "GET":
            return (
                httplib.OK,
                self.fixtures.load("ex_list_stacks.json"),
                {},
                httplib.responses[httplib.OK],
            )
        else:
            return (
                httplib.OK,
                self.fixtures.load("ex_deploy_stack.json"),
                {},
                httplib.responses[httplib.OK],
            )

    def _v1_environments_1e9(self, method, url, body, headers):
        return (
            httplib.OK,
            self.fixtures.load("ex_deploy_stack.json"),
            {},
            httplib.responses[httplib.OK],
        )

    def _v1_environments_1e10(self, method, url, body, headers):
        return (
            httplib.OK,
            self.fixtures.load("ex_destroy_stack.json"),
            {},
            httplib.responses[httplib.OK],
        )

    def _v1_environments_1e1(self, method, url, body, headers):
        return (
            httplib.OK,
            self.fixtures.load("ex_activate_stack.json"),
            {},
            httplib.responses[httplib.OK],
        )

    def _v1_services(self, method, url, body, headers):
        if "?healthState=healthy" in url:
            return (
                httplib.OK,
                self.fixtures.load("ex_search_services.json"),
                {},
                httplib.responses[httplib.OK],
            )
        elif method == "GET":
            return (
                httplib.OK,
                self.fixtures.load("ex_list_services.json"),
                {},
                httplib.responses[httplib.OK],
            )
        else:
            return (
                httplib.OK,
                self.fixtures.load("ex_deploy_service.json"),
                {},
                httplib.responses[httplib.OK],
            )

    def _v1_services_1s13(self, method, url, body, headers):
        if method == "GET":
            return (
                httplib.OK,
                self.fixtures.load("ex_deploy_service.json"),
                {},
                httplib.responses[httplib.OK],
            )
        elif method == "DELETE":
            return (
                httplib.OK,
                self.fixtures.load("ex_destroy_service.json"),
                {},
                httplib.responses[httplib.OK],
            )

    def _v1_services_1s6(self, method, url, body, headers):
        return (
            httplib.OK,
            self.fixtures.load("ex_activate_service.json"),
            {},
            httplib.responses[httplib.OK],
        )

    def _v1_containers(self, method, url, body, headers):
        if "?state=running" in url:
            return (
                httplib.OK,
                self.fixtures.load("ex_search_containers.json"),
                {},
                httplib.responses[httplib.OK],
            )
        elif method == "POST":
            return (
                httplib.OK,
                self.fixtures.load("deploy_container.json"),
                {},
                httplib.responses[httplib.OK],
            )
        return (
            httplib.OK,
            self.fixtures.load("list_containers.json"),
            {},
            httplib.responses[httplib.OK],
        )

    def _v1_containers_1i31(self, method, url, body, headers):
        if method == "GET":
            return (
                httplib.OK,
                self.fixtures.load("deploy_container.json"),
                {},
                httplib.responses[httplib.OK],
            )
        elif method == "DELETE" or "?action=stop" in url:
            return (
                httplib.OK,
                self.fixtures.load("stop_container.json"),
                {},
                httplib.responses[httplib.OK],
            )
        elif "?action=start" in url:
            return (
                httplib.OK,
                self.fixtures.load("start_container.json"),
                {},
                httplib.responses[httplib.OK],
            )
        else:
            return (
                httplib.OK,
                self.fixtures.load("deploy_container.json"),
                {},
                httplib.responses[httplib.OK],
            )


RancherContainerDriver.connectionCls.conn_class = RancherMockHttp
RancherMockHttp.type = None
RancherMockHttp.use_param = "a"


# --------------------------------------------------------------------------- #
# Test Cases


class RancherContainerDriverInitTestCase(unittest.TestCase):
    """
    Tests for testing the different permutations of the driver initialization
    string.
    """

    def test_full_url_string(self):
        """
        Test a 'full' URL string, which contains a scheme, port, and base path.
        """
        path = "http://myhostname:1234/base"
        driver = RancherContainerDriver(*CONTAINER_PARAMS_RANCHER, host=path)

        self.assertEqual(driver.secure, False)
        self.assertEqual(driver.connection.host, "myhostname")
        self.assertEqual(driver.connection.port, 1234)
        self.assertEqual(driver.baseuri, "/base")

    def test_url_string_no_port(self):
        """
        Test a partial URL string, which contains a scheme, and base path.
        """
        path = "http://myhostname/base"
        driver = RancherContainerDriver(*CONTAINER_PARAMS_RANCHER, host=path, port=1234)

        self.assertEqual(driver.secure, False)
        self.assertEqual(driver.connection.host, "myhostname")
        self.assertEqual(driver.connection.port, 1234)
        self.assertEqual(driver.baseuri, "/base")

    def test_url_string_no_scheme(self):
        """
        Test a partial URL string, which contains a port, and base path.
        """
        path = "myhostname:1234/base"
        driver = RancherContainerDriver(*CONTAINER_PARAMS_RANCHER, host=path)

        self.assertEqual(driver.secure, True)
        self.assertEqual(driver.connection.host, "myhostname")
        self.assertEqual(driver.connection.port, 1234)
        self.assertEqual(driver.baseuri, "/base")

    def test_url_string_no_base_path(self):
        """
        Test a partial URL string, which contains a scheme, and a port.
        """
        path = "http://myhostname:1234"
        driver = RancherContainerDriver(*CONTAINER_PARAMS_RANCHER, host=path)

        self.assertEqual(driver.secure, False)
        self.assertEqual(driver.connection.host, "myhostname")
        self.assertEqual(driver.connection.port, 1234)
        self.assertEqual(driver.baseuri, "/v%s" % driver.version)


class RancherContainerDriverTestCase(unittest.TestCase):
    def setUp(self):
        self.driver = RancherContainerDriver(*CONTAINER_PARAMS_RANCHER)

    # Stacks
    def test_ex_list_stacks(self):
        stacks = self.driver.ex_list_stacks()
        self.assertEqual(len(stacks), 6)
        self.assertEqual(stacks[0]["id"], "1e1")

    def test_ex_deploy_stack(self):
        stack = self.driver.ex_deploy_stack(
            name="newstack", environment={"root_password": "password"}
        )
        self.assertEqual(stack["id"], "1e9")
        self.assertEqual(stack["environment"]["root_password"], "password")

    def test_ex_get_stack(self):
        # also uses ex_deploy_stack.json
        stack = self.driver.ex_get_stack("1e9")
        self.assertEqual(stack["id"], "1e9")
        self.assertEqual(stack["environment"]["root_password"], "password")

    def test_ex_search_stacks(self):
        stacks = self.driver.ex_search_stacks({"healthState": "healthy"})
        self.assertEqual(len(stacks), 6)
        self.assertEqual(stacks[0]["healthState"], "healthy")

    def test_ex_destroy_stack(self):
        response = self.driver.ex_destroy_stack("1e10")
        self.assertEqual(response, True)

    def test_ex_activate_stack(self):
        response = self.driver.ex_activate_stack("1e1")
        self.assertEqual(response, True)

    def test_ex_deactivate_stack(self):
        # also uses ex_activate_stack.json
        response = self.driver.ex_activate_stack("1e1")
        self.assertEqual(response, True)

    def test_ex_list_services(self):
        services = self.driver.ex_list_services()
        self.assertEqual(len(services), 4)
        self.assertEqual(services[0]["id"], "1s1")

    def test_ex_deploy_service(self):
        image = ContainerImage(
            id="hastebin",
            name="hastebin",
            path="rlister/hastebin",
            version="latest",
            driver=None,
        )
        service = self.driver.ex_deploy_service(
            name="newservice",
            environment_id="1e1",
            image=image,
            environment={"root_password": "password"},
        )
        self.assertEqual(service["id"], "1s13")
        self.assertEqual(service["environmentId"], "1e6")
        self.assertEqual(service["launchConfig"]["environment"]["root_password"], "password")
        self.assertEqual(service["launchConfig"]["imageUuid"], "docker:rlister/hastebin:latest")

    def test_ex_get_service(self):
        # also uses ex_deploy_service.json
        service = self.driver.ex_get_service("1s13")
        self.assertEqual(service["id"], "1s13")
        self.assertEqual(service["environmentId"], "1e6")
        self.assertEqual(service["launchConfig"]["environment"]["root_password"], "password")

    def test_ex_search_services(self):
        services = self.driver.ex_search_services({"healthState": "healthy"})
        self.assertEqual(len(services), 2)
        self.assertEqual(services[0]["healthState"], "healthy")

    def test_ex_destroy_service(self):
        # Not sure how to do these with returns in mockhttp
        response = self.driver.ex_destroy_service("1s13")
        self.assertEqual(response, True)

    def test_ex_activate_service(self):
        response = self.driver.ex_activate_service("1s6")
        self.assertEqual(response, True)

    def test_ex_deactivate_service(self):
        # also uses ex_activate_service.json
        response = self.driver.ex_activate_service("1s6")
        self.assertEqual(response, True)

    def test_list_containers(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 2)
        self.assertEqual(containers[0].id, "1i1")

    def test_deploy_container(self):
        container = self.driver.deploy_container(
            name="newcontainer",
            image=ContainerImage(
                id="hastebin",
                name="hastebin",
                path="rlister/hastebin",
                version="latest",
                driver=None,
            ),
            environment={"STORAGE_TYPE": "file"},
            networkMode="managed",
        )
        self.assertEqual(container.id, "1i31")
        self.assertEqual(container.name, "newcontainer")
        self.assertEqual(container.extra["environment"], {"STORAGE_TYPE": "file"})

    def test_get_container(self):
        # also uses ex_deploy_container.json
        container = self.driver.get_container("1i31")
        self.assertEqual(container.id, "1i31")
        self.assertEqual(container.name, "newcontainer")
        self.assertEqual(container.extra["environment"], {"STORAGE_TYPE": "file"})

    def test_start_container(self):
        container = self.driver.get_container("1i31")
        started = container.start()
        self.assertEqual(started.id, "1i31")
        self.assertEqual(started.name, "newcontainer")
        self.assertEqual(started.state, "pending")
        self.assertEqual(started.extra["state"], "starting")

    def test_stop_container(self):
        container = self.driver.get_container("1i31")
        stopped = container.stop()
        self.assertEqual(stopped.id, "1i31")
        self.assertEqual(stopped.name, "newcontainer")
        self.assertEqual(stopped.state, "pending")
        self.assertEqual(stopped.extra["state"], "stopping")

    def test_ex_search_containers(self):
        containers = self.driver.ex_search_containers({"state": "running"})
        self.assertEqual(len(containers), 1)

    def test_destroy_container(self):
        container = self.driver.get_container("1i31")
        destroyed = container.destroy()
        self.assertEqual(destroyed.id, "1i31")
        self.assertEqual(destroyed.name, "newcontainer")
        self.assertEqual(destroyed.state, "pending")
        self.assertEqual(destroyed.extra["state"], "stopping")


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import b, httplib, urlparse, parse_qsl
from libcloud.common.types import MalformedResponseError
from libcloud.common.cloudstack import CloudStackConnection

try:
    import simplejson as json
except ImportError:
    import json


async_delay = 0


class CloudStackMockDriver:
    host = "nonexistent."
    path = "/path"
    async_poll_frequency = 0

    name = "fake"

    async_delay = 0


class CloudStackCommonTest(unittest.TestCase):
    def setUp(self):
        CloudStackConnection.conn_class = CloudStackMockHttp
        self.connection = CloudStackConnection("apikey", "secret", host=CloudStackMockDriver.host)
        self.connection.poll_interval = 0.0
        self.driver = self.connection.driver = CloudStackMockDriver()

    def test_sync_request_bad_response(self):
        self.driver.path = "/bad/response"
        try:
            self.connection._sync_request("fake")
        except Exception as e:
            self.assertTrue(isinstance(e, MalformedResponseError))
            return
        self.assertTrue(False)

    def test_sync_request(self):
        self.driver.path = "/sync"
        self.connection._sync_request("fake")

    def test_async_request_successful(self):
        self.driver.path = "/async/success"
        result = self.connection._async_request("fake")
        self.assertEqual(result, {"fake": "result"})

    def test_async_request_unsuccessful(self):
        self.driver.path = "/async/fail"
        try:
            self.connection._async_request("fake")
        except Exception as e:
            self.assertEqual(CloudStackMockHttp.ERROR_TEXT, str(e))
            return
        self.assertFalse(True)

    def test_async_request_delayed(self):
        global async_delay
        self.driver.path = "/async/delayed"
        async_delay = 2
        self.connection._async_request("fake")
        self.assertEqual(async_delay, 0)

    def test_signature_algorithm(self):
        cases = [
            ({"command": "listVirtualMachines"}, "z/a9Y7J52u48VpqIgiwaGUMCso0="),
            (
                {
                    "command": "deployVirtualMachine",
                    "name": "fred",
                    "displayname": "George",
                    "serviceofferingid": 5,
                    "templateid": 17,
                    "zoneid": 23,
                    "networkids": 42,
                },
                "gHTo7mYmadZ+zluKHzlEKb1i/QU=",
            ),
            (
                {
                    "command": "deployVirtualMachine",
                    "name": "fred",
                    "displayname": "George+Ringo",
                    "serviceofferingid": 5,
                    "templateid": 17,
                    "zoneid": 23,
                    "networkids": 42,
                },
                "tAgfrreI1ZvWlWLClD3gu4+aKv4=",
            ),
        ]

        connection = CloudStackConnection("fnord", "abracadabra")
        for case in cases:
            params = connection.add_default_params(case[0])
            self.assertEqual(connection._make_signature(params), b(case[1]))


class CloudStackMockHttp(MockHttp, unittest.TestCase):
    ERROR_TEXT = "ERROR TEXT"

    def _response(self, status, result, response):
        return (status, json.dumps(result), {}, response)

    def _check_request(self, url):
        url = urlparse.urlparse(url)
        query = dict(parse_qsl(url.query))

        self.assertTrue("apiKey" in query)
        self.assertTrue("command" in query)
        self.assertTrue("response" in query)
        self.assertTrue("signature" in query)

        self.assertTrue(query["response"] == "json")

        return query

    def _bad_response(self, method, url, body, headers):
        self._check_request(url)
        result = {"success": True}
        return self._response(httplib.OK, result, httplib.responses[httplib.OK])

    def _sync(self, method, url, body, headers):
        query = self._check_request(url)
        result = {query["command"].lower() + "response": {}}
        return self._response(httplib.OK, result, httplib.responses[httplib.OK])

    def _async_success(self, method, url, body, headers):
        query = self._check_request(url)
        if query["command"].lower() == "queryasyncjobresult":
            self.assertEqual(query["jobid"], "42")
            result = {
                query["command"].lower()
                + "response": {"jobstatus": 1, "jobresult": {"fake": "result"}}
            }
        else:
            result = {query["command"].lower() + "response": {"jobid": "42"}}
        return self._response(httplib.OK, result, httplib.responses[httplib.OK])

    def _async_fail(self, method, url, body, headers):
        query = self._check_request(url)
        if query["command"].lower() == "queryasyncjobresult":
            self.assertEqual(query["jobid"], "42")
            result = {
                query["command"].lower()
                + "response": {
                    "jobstatus": 2,
                    "jobresult": {"errortext": self.ERROR_TEXT},
                }
            }
        else:
            result = {query["command"].lower() + "response": {"jobid": "42"}}
        return self._response(httplib.OK, result, httplib.responses[httplib.OK])

    def _async_delayed(self, method, url, body, headers):
        global async_delay

        query = self._check_request(url)
        if query["command"].lower() == "queryasyncjobresult":
            self.assertEqual(query["jobid"], "42")
            if async_delay == 0:
                result = {
                    query["command"].lower()
                    + "response": {"jobstatus": 1, "jobresult": {"fake": "result"}}
                }
            else:
                result = {query["command"].lower() + "response": {"jobstatus": 0}}
                async_delay -= 1
        else:
            result = {query["command"].lower() + "response": {"jobid": "42"}}
        return self._response(httplib.OK, result, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Abiquo Compute Driver

The driver implements the compute Abiquo functionality for the Abiquo API.
This version is compatible with the following versions of Abiquo:

    * Abiquo 3.4 (http://wiki.abiquo.com/display/ABI34/The+Abiquo+API)
"""

from libcloud.utils.py3 import ET, tostring
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.common.abiquo import AbiquoResponse, AbiquoConnection, get_href
from libcloud.compute.types import Provider, LibcloudError


class AbiquoNodeDriver(NodeDriver):
    """
    Implements the :class:`NodeDriver`'s for the Abiquo Compute Provider
    """

    type = Provider.ABIQUO
    name = "Abiquo"
    website = "http://www.abiquo.com/"
    connectionCls = AbiquoConnection
    timeout = 2000  # some images take a lot of time!

    # Media Types
    NODES_MIME_TYPE = "application/vnd.abiquo.virtualmachines+xml"
    NODE_MIME_TYPE = "application/vnd.abiquo.virtualmachine+xml"
    VAPPS_MIME_TYPE = "application/vnd.abiquo.virtualappliances+xml"
    VAPP_MIME_TYPE = "application/vnd.abiquo.virtualappliance+xml"
    VM_TASK_MIME_TYPE = "application/vnd.abiquo.virtualmachinetask+xml"
    USER_MIME_TYPE = "application/vnd.abiquo.user+xml"
    ENT_MIME_TYPE = "application/vnd.abiquo.enterprise+xml"
    VDCS_MIME_TYPE = "application/vnd.abiquo.virtualdatacenters+xml"
    VDC_MIME_TYPE = "application/vnd.abiquo.virtualdatacenter+xml"
    DCS_MIME_TYPE = "application/vnd.abiquo.datacenters+xml"
    VMTPLS_MIME_TYPE = "application/vnd.abiquo.virtualmachinetemplates+xml"
    VMTPL_MIME_TYPE = "application/vnd.abiquo.virtualmachinetemplate+xml"
    NICS_MIME_TYPE = "application/vnd.abiquo.nics+xml"
    DCRS_MIME_TYPE = "application/vnd.abiquo.datacenterrepositories+xml"
    DCR_MIME_TYPE = "application/vnd.abiquo.datacenterrepository+xml"
    AR_MIME_TYPE = "application/vnd.abiquo.acceptedrequest+xml"

    # Others constants
    GIGABYTE = 1073741824

    def __init__(self, user_id, secret, endpoint, **kwargs):
        """
        Initializes Abiquo Driver

        Initializes the :class:`NodeDriver` object and populate the cache.

        :param       user_id: identifier of Abiquo user (required)
        :type        user_id: ``str``
        :param       secret: password of the Abiquo user (required)
        :type        secret: ``str``
        :param       endpoint: Abiquo API endpoint (required)
        :type        endpoint: ``str`` that can be parsed as URL
        """
        self.endpoint = endpoint
        super().__init__(key=user_id, secret=secret, secure=False, host=None, port=None, **kwargs)
        self.ex_populate_cache()

    def create_node(self, image, name=None, size=None, location=None, ex_group_name=None):
        """
        Create a new node instance in Abiquo

        All the :class:`Node`s need to be defined inside a VirtualAppliance
        (called :class:`NodeGroup` here). If there is no group name defined,
        'libcloud' name will be used instead.

        This method wraps these Abiquo actions:

            1. Create a group if it does not exist.
            2. Register a new node in the group.
            3. Deploy the node and boot it.
            4. Retrieves it again to get schedule-time attributes (such as ips
               and vnc ports).

        The rest of the driver methods has been created in a way that, if any
        of these actions fail, the user can not reach an inconsistent state

        :keyword    name:   The name for this new node (required)
        :type       name:   ``str``

        :keyword    size:   The size of resources allocated to this node.
        :type       size:   :class:`NodeSize`

        :keyword    image:  OS Image to boot on node. (required)
        :type       image:  :class:`NodeImage`

        :keyword    location: Which data center to create a node in. If empty,
                              undefined behavior will be selected. (optional)
        :type       location: :class:`NodeLocation`

        :keyword   ex_group_name:  Which group this node belongs to. If empty,
                                   it will be created into 'libcloud' group. If
                                   it does not found any group in the target
                                   location (random location if you have not
                                   set the parameter), then it will create a
                                   new group with this name.
        :type     ex_group_name:  c{str}

        :return:               The newly created node.
        :rtype:                :class:`Node`
        """
        # Define the location
        # To be clear:
        #     'xml_loc' is the xml element we navigate into (we need links)
        #     'loc' is the :class:`NodeLocation` entity
        xml_loc, loc = self._define_create_node_location(image=image, location=location)

        # Define the Group
        group = self._define_create_node_group(xml_loc, loc, ex_group_name)

        # Register the Node
        vm = self._define_create_node_node(group, name=name, size=size, image=image)

        # Execute the 'create' in hypervisor action
        self._deploy_remote(vm)

        # Retrieve it again, to get some schedule-time defined values
        edit_vm = get_href(vm, "edit")
        headers = {"Accept": self.NODE_MIME_TYPE}
        vm = self.connection.request(edit_vm, headers=headers).object
        return self._to_node(vm, self)

    def destroy_node(self, node):
        """
        Destroy a node

        Depending on the provider, this may destroy all data associated with
        the node, including backups.

        :param node: The node to be destroyed
        :type node: :class:`Node`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``
        """

        # Refresh node state
        headers = {"Accept": self.NODE_MIME_TYPE}
        e_vm = self.connection.request(node.extra["uri_id"], headers=headers).object

        state = e_vm.findtext("state")

        if state in ["ALLOCATED", "CONFIGURED", "LOCKED", "UNKNOWN"]:
            raise LibcloudError("Invalid Node state", self)

        if state != "NOT_ALLOCATED":
            # prepare the element that forces the undeploy
            vm_task = ET.Element("virtualmachinetask")
            force_undeploy = ET.SubElement(vm_task, "forceUndeploy")
            force_undeploy.text = "True"
            # Set the URI
            destroy_uri = node.extra["uri_id"] + "/action/undeploy"
            # Prepare the headers
            headers = {
                "Accept": self.AR_MIME_TYPE,
                "Content-type": self.VM_TASK_MIME_TYPE,
            }
            res = self.connection.async_request(
                action=destroy_uri,
                method="POST",
                data=tostring(vm_task),
                headers=headers,
            )

        # pylint: disable=maybe-no-member
        if state == "NOT_ALLOCATED" or res.async_success():
            # pylint: enable=maybe-no-member
            self.connection.request(action=node.extra["uri_id"], method="DELETE")
            return True
        else:
            return False

    def ex_run_node(self, node):
        """
        Runs a node

        Here there is a bit difference between Abiquo states and libcloud
        states, so this method is created to have better compatibility. In
        libcloud, if the node is not running, then it does not exist (avoiding
        UNKNOWN and temporal states). In Abiquo, you can define a node, and
        then deploy it.

        If the node is in :class:`NodeState.TERMINATED` libcloud's state and in
        'NOT_DEPLOYED' Abiquo state, there is a way to run and recover it
        for libcloud using this method. There is no way to reach this state
        if you are using only libcloud, but you may have used another Abiquo
        client and now you want to recover your node to be used by libcloud.

        :param node: The node to run
        :type node: :class:`Node`

        :return: The node itself, but with the new state
        :rtype: :class:`Node`
        """
        # Refresh node state
        e_vm = self.connection.request(node.extra["uri_id"]).object
        state = e_vm.findtext("state")

        if state != "NOT_ALLOCATED":
            raise LibcloudError("Invalid Node state", self)

        # --------------------------------------------------------
        #     Deploy the Node
        # --------------------------------------------------------
        self._deploy_remote(e_vm)

        # --------------------------------------------------------
        #     Retrieve it again, to get some schedule-defined
        #     values.
        # --------------------------------------------------------
        edit_vm = get_href(e_vm, "edit")
        headers = {"Accept": self.NODE_MIME_TYPE}
        e_vm = self.connection.request(edit_vm, headers=headers).object
        return self._to_node(e_vm, self)

    def ex_populate_cache(self):
        """
        Populate the cache.

        For each connection, it is good to store some objects that will be
        useful for further requests, such as the 'user' and the 'enterprise'
        objects.

        Executes the 'login' resource after setting the connection parameters
        and, if the execution is successful, it sets the 'user' object into
        cache. After that, it also requests for the 'enterprise' and
        'locations' data.

        List of locations should remain the same for a single libcloud
        connection. However, this method is public and you are able to
        refresh the list of locations any time.
        """

        user_headers = {"Accept": self.USER_MIME_TYPE}
        user = self.connection.request("/login", headers=user_headers).object
        self.connection.cache["user"] = user
        e_ent = get_href(self.connection.cache["user"], "enterprise")
        ent_headers = {"Accept": self.ENT_MIME_TYPE}
        ent = self.connection.request(e_ent, headers=ent_headers).object
        self.connection.cache["enterprise"] = ent

        vdcs_headers = {"Accept": self.VDCS_MIME_TYPE}
        uri_vdcs = "/cloud/virtualdatacenters"
        e_vdcs = self.connection.request(uri_vdcs, headers=vdcs_headers).object

        params = {"idEnterprise": self._get_enterprise_id()}

        dcs_headers = {"Accept": self.DCS_MIME_TYPE}
        e_dcs = self.connection.request(
            "/admin/datacenters", headers=dcs_headers, params=params
        ).object
        dc_dict = {}
        for dc in e_dcs.findall("datacenter"):
            key = get_href(dc, "self")
            dc_dict[key] = dc

        # Populate locations name cache
        self.connection.cache["locations"] = {}
        for e_vdc in e_vdcs.findall("virtualDatacenter"):
            loc = get_href(e_vdc, "location")
            if loc is not None:
                self.connection.cache["locations"][loc] = get_href(e_vdc, "edit")

    def ex_create_group(self, name, location=None):
        """
        Create an empty group.

        You can specify the location as well.

        :param     group:     name of the group (required)
        :type      group:     ``str``

        :param     location: location were to create the group
        :type      location: :class:`NodeLocation`

        :returns:            the created group
        :rtype:              :class:`NodeGroup`
        """
        # prepare the element
        vapp = ET.Element("virtualAppliance")
        vapp_name = ET.SubElement(vapp, "name")
        vapp_name.text = name

        if location is None:
            location = self.list_locations()[0]
        elif location not in self.list_locations():
            raise LibcloudError("Location does not exist")

        link_vdc = self.connection.cache["locations"][location]
        hdr_vdc = {"Accept": self.VDC_MIME_TYPE}
        e_vdc = self.connection.request(link_vdc, headers=hdr_vdc).object

        creation_link = get_href(e_vdc, "virtualappliances")
        headers = {"Accept": self.VAPP_MIME_TYPE, "Content-type": self.VAPP_MIME_TYPE}
        vapp = self.connection.request(
            creation_link, data=tostring(vapp), headers=headers, method="POST"
        ).object

        uri_vapp = get_href(vapp, "edit")

        return NodeGroup(self, vapp.findtext("name"), uri=uri_vapp)

    def ex_destroy_group(self, group):
        """
        Destroy a group.

        Be careful! Destroying a group means destroying all the :class:`Node`
        instances there and the group itself!

        If there is currently any action over any :class:`Node` of the
        :class:`NodeGroup`, then the method will raise an exception.

        :param     name: The group (required)
        :type      name: :class:`NodeGroup`

        :return:         If the group was destroyed successfully
        :rtype:          ``bool``
        """
        # Refresh group state
        e_group = self.connection.request(group.uri).object
        state = e_group.findtext("state")

        if state not in ["NOT_DEPLOYED", "DEPLOYED"]:
            error = "Can not destroy group because of current state"
            raise LibcloudError(error, self)

        if state == "DEPLOYED":
            # prepare the element that forces the undeploy
            vm_task = ET.Element("virtualmachinetask")
            force_undeploy = ET.SubElement(vm_task, "forceUndeploy")
            force_undeploy.text = "True"

            # Set the URI
            undeploy_uri = group.uri + "/action/undeploy"

            # Prepare the headers
            headers = {
                "Accept": self.AR_MIME_TYPE,
                "Content-type": self.VM_TASK_MIME_TYPE,
            }
            res = self.connection.async_request(
                action=undeploy_uri,
                method="POST",
                data=tostring(vm_task),
                headers=headers,
            )

        # pylint: disable=maybe-no-member
        if state == "NOT_DEPLOYED" or res.async_success():
            # pylint: enable=maybe-no-member
            # The node is no longer deployed. Unregister it.
            self.connection.request(action=group.uri, method="DELETE")
            return True
        else:
            return False

    def ex_list_groups(self, location=None):
        """
        List all groups.

        :param location: filter the groups by location (optional)
        :type  location: a :class:`NodeLocation` instance.

        :return:         the list of :class:`NodeGroup`
        """
        groups = []
        for vdc in self._get_locations(location):
            link_vdc = self.connection.cache["locations"][vdc]
            hdr_vdc = {"Accept": self.VDC_MIME_TYPE}
            e_vdc = self.connection.request(link_vdc, headers=hdr_vdc).object
            apps_link = get_href(e_vdc, "virtualappliances")
            hdr_vapps = {"Accept": self.VAPPS_MIME_TYPE}
            vapps = self.connection.request(apps_link, headers=hdr_vapps).object
            for vapp in vapps.findall("virtualAppliance"):
                nodes = []
                vms_link = get_href(vapp, "virtualmachines")
                headers = {"Accept": self.NODES_MIME_TYPE}
                vms = self.connection.request(vms_link, headers=headers).object
                for vm in vms.findall("virtualMachine"):
                    nodes.append(self._to_node(vm, self))
                group = NodeGroup(self, vapp.findtext("name"), nodes, get_href(vapp, "edit"))
                groups.append(group)

        return groups

    def list_images(self, location=None):
        """
        List images on Abiquo Repositories

        :keyword location: The location to list images for.
        :type    location: :class:`NodeLocation`

        :return:           list of node image objects
        :rtype:            ``list`` of :class:`NodeImage`
        """
        enterprise_id = self._get_enterprise_id()
        uri = "/admin/enterprises/%s/datacenterrepositories/" % (enterprise_id)
        repos_hdr = {"Accept": self.DCRS_MIME_TYPE}
        repos = self.connection.request(uri, headers=repos_hdr).object

        images = []
        for repo in repos.findall("datacenterRepository"):
            # filter by location. Skips when the name of the location
            # is different from the 'datacenterRepository' element
            for vdc in self._get_locations(location):
                # Check if the virtual datacenter belongs to this repo
                link_vdc = self.connection.cache["locations"][vdc]
                hdr_vdc = {"Accept": self.VDC_MIME_TYPE}
                e_vdc = self.connection.request(link_vdc, headers=hdr_vdc).object
                dc_link_vdc = get_href(e_vdc, "location")
                dc_link_repo = get_href(repo, "datacenter")

                if dc_link_vdc.split("/")[-1] == dc_link_repo.split("/")[-1]:
                    # Filter the template in case we don't have it yet
                    url_templates = get_href(repo, "virtualmachinetemplates")
                    hypervisor_type = e_vdc.findtext("hypervisorType")
                    params = {"hypervisorTypeName": hypervisor_type}
                    headers = {"Accept": self.VMTPLS_MIME_TYPE}
                    templates = self.connection.request(
                        url_templates, params, headers=headers
                    ).object
                    for templ in templates.findall("virtualMachineTemplate"):
                        # Avoid duplicated templates
                        id_template = templ.findtext("id")
                        ids = [image.id for image in images]
                        if id_template not in ids:
                            images.append(self._to_nodeimage(templ, self, get_href(repo, "edit")))

        return images

    def list_locations(self):
        """
        Return list of locations where the user has access to.

        :return: the list of :class:`NodeLocation` available for the current
                 user
        :rtype:  ``list`` of :class:`NodeLocation`
        """
        return list(self.connection.cache["locations"].keys())

    def list_nodes(self, location=None):
        """
        List all nodes.

        :param location: Filter the groups by location (optional)
        :type  location: a :class:`NodeLocation` instance.

        :return:  List of node objects
        :rtype: ``list`` of :class:`Node`
        """
        nodes = []

        for group in self.ex_list_groups(location):
            nodes.extend(group.nodes)

        return nodes

    def list_sizes(self, location=None):
        """
        List sizes on a provider.

        Abiquo does not work with sizes. However, this method
        returns a list of predefined ones (copied from :class:`DummyNodeDriver`
        but without price neither bandwidth) to help the users to create their
        own.

        If you call the method :class:`AbiquoNodeDriver.create_node` with the
        size informed, it will just override the 'ram' value of the 'image'
        template. So it is no too much useful work with sizes...

        :return: The list of sizes
        :rtype:  ``list`` of :class:`NodeSizes`
        """
        return [
            NodeSize(
                id=1,
                name="Small",
                ram=128,
                disk=4,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            NodeSize(
                id=2,
                name="Medium",
                ram=512,
                disk=16,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            NodeSize(
                id=3,
                name="Big",
                ram=4096,
                disk=32,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            NodeSize(
                id=4,
                name="XXL Big",
                ram=4096 * 2,
                disk=32 * 4,
                bandwidth=None,
                price=None,
                driver=self,
            ),
        ]

    def reboot_node(self, node):
        """
        Reboot a node.

        :param node: The node to be rebooted
        :type node: :class:`Node`

        :return: True if the reboot was successful, otherwise False
        :rtype: ``bool``
        """
        reboot_uri = node.extra["uri_id"] + "/action/reset"
        reboot_hdr = {"Accept": self.AR_MIME_TYPE}
        res = self.connection.async_request(action=reboot_uri, method="POST", headers=reboot_hdr)
        return res.async_success()  # pylint: disable=maybe-no-member

    # -------------------------
    # Extension methods
    # -------------------------

    def _ex_connection_class_kwargs(self):
        """
        Set the endpoint as an extra :class:`AbiquoConnection` argument.

        According to Connection code, the "url" argument should be
        parsed properly to connection.

        :return: ``dict`` of :class:`AbiquoConnection` input arguments
        """

        return {"url": self.endpoint}

    def _deploy_remote(self, e_vm):
        """
        Asynchronous call to create the node.
        """
        # --------------------------------------------------------
        #     Deploy the Node
        # --------------------------------------------------------
        # prepare the element that forces the deploy
        vm_task = ET.Element("virtualmachinetask")
        force_deploy = ET.SubElement(vm_task, "forceEnterpriseSoftLimits")
        force_deploy.text = "True"

        # Prepare the headers
        headers = {"Accept": self.AR_MIME_TYPE, "Content-type": self.VM_TASK_MIME_TYPE}
        link_deploy = get_href(e_vm, "deploy")
        res = self.connection.async_request(
            action=link_deploy, method="POST", data=tostring(vm_task), headers=headers
        )
        if not res.async_success():  # pylint: disable=maybe-no-member
            raise LibcloudError("Could not run the node", self)

    def _to_location(self, vdc, dc, driver):
        """
        Generates the :class:`NodeLocation` class.
        """
        identifier = vdc.findtext("id")
        name = vdc.findtext("name")
        country = dc.findtext("name")
        return NodeLocation(identifier, name, country, driver)

    def _to_node(self, vm, driver):
        """
        Generates the :class:`Node` class.
        """
        identifier = vm.findtext("id")
        name = vm.findtext("label")
        state = AbiquoResponse.NODE_STATE_MAP[vm.findtext("state")]

        link_image = get_href(vm, "virtualmachinetemplate")
        link_hdr = {"Accept": self.VMTPL_MIME_TYPE}
        image_element = self.connection.request(link_image, headers=link_hdr).object
        repo_link = get_href(image_element, "datacenterrepository")
        image = self._to_nodeimage(image_element, self, repo_link)

        # Fill the 'ips' data
        private_ips = []
        public_ips = []
        nics_hdr = {"Accept": self.NICS_MIME_TYPE}
        nics_element = self.connection.request(get_href(vm, "nics"), headers=nics_hdr).object
        for nic in nics_element.findall("nic"):
            ip = nic.findtext("ip")
            for link in nic.findall("link"):
                rel = link.attrib["rel"]
                if rel == "privatenetwork":
                    private_ips.append(ip)
                elif rel in ["publicnetwork", "externalnetwork", "unmanagednetwork"]:
                    public_ips.append(ip)

        extra = {"uri_id": get_href(vm, "edit")}

        if vm.find("vdrpIp") is not None:
            extra["vdrp_ip"] = vm.findtext("vdrpIP")
            extra["vdrp_port"] = vm.findtext("vdrpPort")

        return Node(
            identifier,
            name,
            state,
            public_ips,
            private_ips,
            driver,
            image=image,
            extra=extra,
        )

    def _to_nodeimage(self, template, driver, repo):
        """
        Generates the :class:`NodeImage` class.
        """
        identifier = template.findtext("id")
        name = template.findtext("name")
        url = get_href(template, "edit")
        hdreqd = template.findtext("hdRequired")
        extra = {"repo": repo, "url": url, "hdrequired": hdreqd}
        return NodeImage(identifier, name, driver, extra)

    def _get_locations(self, location=None):
        """
        Returns the locations as a generator.
        """
        if location is not None:
            yield location
        else:
            yield from self.list_locations()

    def _get_enterprise_id(self):
        """
        Returns the identifier of the logged user's enterprise.
        """
        return self.connection.cache["enterprise"].findtext("id")

    def _define_create_node_location(self, image, location):
        """
        Search for a location where to create the node.

        Based on 'create_node' **kwargs argument, decide in which
        location will be created.
        """
        # First, get image location
        if not image:
            error = "'image' parameter is mandatory"
            raise LibcloudError(error, self)

        # Get the location argument
        if location:
            if location not in self.list_locations():
                raise LibcloudError("Location does not exist")

        # Check if the image is compatible with any of the locations or
        # the input location
        loc = None
        target_loc = None
        for candidate_loc in self._get_locations(location):
            link_vdc = self.connection.cache["locations"][candidate_loc]
            hdr_vdc = {"Accept": self.VDC_MIME_TYPE}
            e_vdc = self.connection.request(link_vdc, headers=hdr_vdc).object
            for img in self.list_images(candidate_loc):
                if img.id == image.id:
                    loc = e_vdc
                    target_loc = candidate_loc
                    break

        if loc is None:
            error = "The image can not be used in any location"
            raise LibcloudError(error, self)

        return loc, target_loc

    def _define_create_node_group(self, xml_loc, loc, group_name=None):
        """
        Search for a group where to create the node.

        If we can not find any group, create it into argument 'location'
        """
        if not group_name:
            group_name = NodeGroup.DEFAULT_GROUP_NAME

        # We search if the group is already defined into the location
        groups_link = get_href(xml_loc, "virtualappliances")
        groups_hdr = {"Accept": self.VAPPS_MIME_TYPE}
        vapps_element = self.connection.request(groups_link, headers=groups_hdr).object
        target_group = None
        for vapp in vapps_element.findall("virtualAppliance"):
            if vapp.findtext("name") == group_name:
                uri_vapp = get_href(vapp, "edit")
                return NodeGroup(self, vapp.findtext("name"), uri=uri_vapp)

        # target group not found: create it. Since it is an extension of
        # the basic 'libcloud' functionality, we try to be as flexible as
        # possible.
        if target_group is None:
            return self.ex_create_group(group_name, loc)

    def _define_create_node_node(self, group, name=None, size=None, image=None):
        """
        Defines the node before to create.

        In Abiquo, you first need to 'register' or 'define' the node in
        the API before to create it into the target hypervisor.
        """
        vm = ET.Element("virtualMachine")
        if name:
            vmname = ET.SubElement(vm, "label")
            vmname.text = name
        attrib = {
            "type": self.VMTPL_MIME_TYPE,
            "rel": "virtualmachinetemplate",
            "href": image.extra["url"],
        }
        ET.SubElement(vm, "link", attrib=attrib)
        headers = {"Accept": self.NODE_MIME_TYPE, "Content-type": self.NODE_MIME_TYPE}

        if size:
            # Override the 'NodeSize' data
            ram = ET.SubElement(vm, "ram")
            ram.text = str(size.ram)

        # Create the virtual machine
        nodes_link = group.uri + "/virtualmachines"
        vm = self.connection.request(
            nodes_link, data=tostring(vm), headers=headers, method="POST"
        ).object
        edit_vm = get_href(vm, "edit")
        headers = {"Accept": self.NODE_MIME_TYPE}

        return self.connection.request(edit_vm, headers=headers).object


class NodeGroup:
    """
    Group of virtual machines that can be managed together

    All :class:`Node`s in Abiquo must be defined inside a Virtual Appliance.
    We offer a way to handle virtual appliances (called NodeGroup to
    maintain some kind of name conventions here) inside the
    :class:`AbiquoNodeDriver` without breaking compatibility of the rest of
    libcloud API.

    If the user does not want to handle groups, all the virtual machines
    will be created inside a group named 'libcloud'
    """

    DEFAULT_GROUP_NAME = "libcloud"

    def __init__(self, driver, name=DEFAULT_GROUP_NAME, nodes=[], uri=""):
        """
        Initialize a new group object.
        """
        self.driver = driver
        self.name = name
        self.nodes = nodes
        self.uri = uri

    def __repr__(self):
        return ("<NodeGroup: name=%s, nodes=[%s] >") % (
            self.name,
            ",".join(map(str, self.nodes)),
        )

    def destroy(self):
        """
        Destroys the group delegating the execution to
        :class:`AbiquoNodeDriver`.
        """
        return self.driver.ex_destroy_group(self)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib, xmlrpclib
from libcloud.test.secrets import VCL_PARAMS
from libcloud.compute.types import NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.vcl import VCLNodeDriver as VCL


class VCLTests(unittest.TestCase):
    def setUp(self):
        VCL.connectionCls.conn_class = VCLMockHttp
        VCLMockHttp.type = None
        self.driver = VCL(*VCL_PARAMS)

    def test_list_nodes(self):
        node = self.driver.list_nodes(ipaddr="192.168.1.1")[0]
        self.assertEqual(node.name, "CentOS 5.4 Base (32 bit VM)")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.extra["pass"], "ehkNGW")

    def test_list_images(self):
        images = self.driver.list_images()
        image = images[0]
        self.assertEqual(image.id, "8")

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 1)

    def test_create_node(self):
        image = self.driver.list_images()[0]
        node = self.driver.create_node(image=image)
        self.assertEqual(node.id, "51")

    def test_destroy_node(self):
        node = self.driver.list_nodes(ipaddr="192.168.1.1")[0]
        self.assertTrue(self.driver.destroy_node(node))

    def test_ex_update_node_access(self):
        node = self.driver.list_nodes(ipaddr="192.168.1.1")[0]
        node = self.driver.ex_update_node_access(node, ipaddr="192.168.1.2")
        self.assertEqual(node.name, "CentOS 5.4 Base (32 bit VM)")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.extra["pass"], "ehkNGW")

    def test_ex_extend_request_time(self):
        node = self.driver.list_nodes(ipaddr="192.168.1.1")[0]
        self.assertTrue(self.driver.ex_extend_request_time(node, 60))

    def test_ex_get_request_end_time(self):
        node = self.driver.list_nodes(ipaddr="192.168.1.1")[0]
        self.assertEqual(self.driver.ex_get_request_end_time(node), 1334168100)


class VCLMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("vcl")

    def _get_method_name(self, type, use_param, qs, path):
        return "_xmlrpc"

    def _xmlrpc(self, method, url, body, headers):
        params, meth_name = xmlrpclib.loads(body)
        if self.type:
            meth_name = "{}_{}".format(meth_name, self.type)
        return getattr(self, meth_name)(method, url, body, headers)

    def XMLRPCgetImages(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCgetImages.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCextendRequest(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCextendRequest.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCgetRequestIds(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCgetRequestIds.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCgetRequestStatus(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCgetRequestStatus.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCendRequest(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCendRequest.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCaddRequest(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCaddRequest.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def XMLRPCgetRequestConnectData(self, method, url, body, headers):
        body = self.fixtures.load("XMLRPCgetRequestConnectData.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["CloudFlareDNSDriver"]

import json
import itertools

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    Provider,
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.utils.misc import reverse_dict, merge_valid_keys
from libcloud.common.base import JsonResponse, ConnectionKey, ConnectionUserAndKey
from libcloud.common.types import LibcloudError, InvalidCredsError

API_HOST = "api.cloudflare.com"
API_BASE = "/client/v4"

CLOUDFLARE_TO_LIBCLOUD_ZONE_TYPE = {
    "full": "master",
    "partial": "slave",
}

LIBCLOUD_TO_CLOUDFLARE_ZONE_TYPE = reverse_dict(CLOUDFLARE_TO_LIBCLOUD_ZONE_TYPE)

ZONE_EXTRA_ATTRIBUTES = {
    "development_mode",
    "original_name_servers",
    "original_registrar",
    "original_dnshost",
    "created_on",
    "modified_on",
    "activated_on",
    "owner",
    "account",
    "permissions",
    "plan",
    "plan_pending",
    "status",
    "paused",
    "name_servers",
}

ZONE_UPDATE_ATTRIBUTES = {
    "paused",
    "vanity_name_servers",
    "plan",
}

ZONE_CREATE_ATTRIBUTES = {
    "jump_start",
}

RECORD_EXTRA_ATTRIBUTES = {
    "proxiable",
    "proxied",
    "locked",
    "created_on",
    "modified_on",
    "data",
    "priority",
}

RECORD_CREATE_ATTRIBUTES = {
    "ttl",
    "priority",
    "proxied",
}

RECORD_UPDATE_ATTRIBUTES = {
    "ttl",
    "proxied",
}


class CloudFlareDNSResponse(JsonResponse):
    exceptions = {
        9103: (InvalidCredsError, []),
        1001: (ZoneDoesNotExistError, ["zone_id"]),
        1061: (ZoneAlreadyExistsError, ["zone_id"]),
        1002: (RecordDoesNotExistError, ["record_id"]),
        81053: (RecordAlreadyExistsError, ["record_id"]),
        # 81057: The record already exists.
        81057: (RecordAlreadyExistsError, []),
        # 81058: A record with those settings already exists.
        81058: (RecordAlreadyExistsError, ["record_id"]),
    }

    def success(self):
        body = self.parse_body()

        is_success = body.get("success", False)

        return is_success

    def parse_error(self):
        body = self.parse_body()

        errors = body.get("errors", [])

        for error in errors:
            error_chain = error.get("error_chain", [])

            error_chain_errors = []
            for chain_error in error_chain:
                error_chain_errors.append(
                    "%s: %s"
                    % (
                        chain_error.get("code", "unknown"),
                        chain_error.get("message", ""),
                    )
                )

            try:
                exception_class, context = self.exceptions[error["code"]]
            except KeyError:
                exception_class, context = LibcloudError, []

            kwargs = {
                "value": "{}: {} (error chain: {})".format(
                    error["code"], error["message"], ", ".join(error_chain_errors)
                ),
                "driver": self.connection.driver,
            }

            if error["code"] == 81057:
                # Record id is not available when creating a record and not
                # updating it
                kwargs["record_id"] = "unknown"

            merge_valid_keys(kwargs, context, self.connection.context)

            raise exception_class(**kwargs)


class BaseDNSConnection:
    host = API_HOST
    secure = True
    responseCls = CloudFlareDNSResponse

    def encode_data(self, data):
        return json.dumps(data)


class GlobalAPIKeyDNSConnection(BaseDNSConnection, ConnectionUserAndKey):
    def add_default_headers(self, headers):
        headers["Content-Type"] = "application/json"
        headers["X-Auth-Email"] = self.user_id
        headers["X-Auth-Key"] = self.key

        return headers


class TokenDNSConnection(BaseDNSConnection, ConnectionKey):
    def add_default_headers(self, headers):
        headers["Content-Type"] = "application/json"
        headers["Authorization"] = "Bearer %s" % self.key

        return headers


class CloudFlareDNSDriver(DNSDriver):
    type = Provider.CLOUDFLARE
    name = "CloudFlare DNS"
    website = "https://www.cloudflare.com"
    connectionCls = GlobalAPIKeyDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.TXT: "TXT",
        RecordType.SPF: "SPF",
        RecordType.NS: "NS",
        RecordType.SRV: "SRV",
        RecordType.URL: "LOC",
    }

    ZONES_PAGE_SIZE = 50
    RECORDS_PAGE_SIZE = 100
    MEMBERSHIPS_PAGE_SIZE = 50

    def __init__(self, key, secret=None, secure=True, host=None, port=None, **kwargs):
        if secret is None:
            self.connectionCls = TokenDNSConnection

        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port, **kwargs)

    def iterate_zones(self):
        def _iterate_zones(params):
            url = "{}/zones".format(API_BASE)

            response = self.connection.request(url, params=params)

            items = response.object["result"]
            zones = [self._to_zone(item) for item in items]

            return response, zones

        return self._paginate(_iterate_zones, self.ZONES_PAGE_SIZE)

    def iterate_records(self, zone):
        def _iterate_records(params):
            url = "{}/zones/{}/dns_records".format(API_BASE, zone.id)

            self.connection.set_context({"zone_id": zone.id})
            response = self.connection.request(url, params=params)

            items = response.object["result"]
            records = [self._to_record(zone, item) for item in items]

            return response, records

        return self._paginate(_iterate_records, self.RECORDS_PAGE_SIZE)

    def get_zone(self, zone_id):
        url = "{}/zones/{}".format(API_BASE, zone_id)

        self.connection.set_context({"zone_id": zone_id})
        response = self.connection.request(url)

        item = response.object["result"]
        zone = self._to_zone(item)

        return zone

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        @inherits: :class:`DNSDriver.create_zone`

        Note that for users who have more than one account membership,
        the id of the account in which to create the zone must be
        specified via the ``extra`` key ``account``.

        Note that for ``extra`` zone properties, only the ones specified in
        ``ZONE_CREATE_ATTRIBUTES`` can be set at creation time. Additionally,
        setting the ``ttl` property is not supported.
        """
        extra = extra or {}

        account = extra.get("account")
        if account is None:
            memberships = self.ex_get_user_account_memberships()
            memberships = list(itertools.islice(memberships, 2))

            if len(memberships) != 1:
                raise ValueError("must specify account for zone")

            account = memberships[0]["account"]["id"]

        url = "{}/zones".format(API_BASE)

        body = {
            "name": domain,
            "account": {"id": account},
            "type": LIBCLOUD_TO_CLOUDFLARE_ZONE_TYPE[type],
        }

        merge_valid_keys(body, ZONE_CREATE_ATTRIBUTES, extra)

        response = self.connection.request(url, data=body, method="POST")

        item = response.object["result"]
        zone = self._to_zone(item)

        return zone

    def update_zone(self, zone, domain, type="master", ttl=None, extra=None):
        """
        @inherits: :class:`DNSDriver.update_zone`

        Note that the ``zone``, ``type`` and ``ttl`` properties can't be
        updated. The only updatable properties are the ``extra`` zone
        properties specified in ``ZONE_UPDATE_ATTRIBUTES``. Only one property
        may be updated at a time. Any non-updatable properties are ignored.
        """
        body = merge_valid_keys({}, ZONE_UPDATE_ATTRIBUTES, extra)
        if len(body) != 1:
            return zone

        url = "{}/zones/{}".format(API_BASE, zone.id)

        self.connection.set_context({"zone_id": zone.id})
        response = self.connection.request(url, data=body, method="PATCH")

        item = response.object["result"]
        zone = self._to_zone(item)

        return zone

    def delete_zone(self, zone):
        url = "{}/zones/{}".format(API_BASE, zone.id)

        self.connection.set_context({"zone_id": zone.id})
        response = self.connection.request(url, method="DELETE")

        item = response.object.get("result", {}).get("id")
        is_deleted = item == zone.id

        return is_deleted

    def get_record(self, zone_id, record_id):
        zone = self.get_zone(zone_id)

        url = "{}/zones/{}/dns_records/{}".format(API_BASE, zone.id, record_id)

        self.connection.set_context({"record_id": record_id})
        response = self.connection.request(url)

        item = response.object["result"]
        record = self._to_record(zone, item)

        return record

    def create_record(self, name, zone, type, data, extra=None):
        """
        @inherits: :class:`DNSDriver.create_record`

        Note that for ``extra`` record properties, only the ones specified in
        ``RECORD_CREATE_ATTRIBUTES`` can be set at creation time. Any
        non-settable properties are ignored.

        NOTE: For CAA RecordType, data needs to be in the following format:
        <flags> <tag> <ca domain name> where the tag can be issue, issuewild
        or iodef.

        For example: 0 issue test.caa.com

        NOTE: For SSHFP RecordType, data need to be in the format:
        <algorithm> <type> <fingerprint>
        """
        url = "{}/zones/{}/dns_records".format(API_BASE, zone.id)

        content, data = self._normalize_record_data_for_api(type=type, data=data)
        body = {"type": type, "name": name, "content": content}

        if data:
            body["data"] = data

        merge_valid_keys(body, RECORD_CREATE_ATTRIBUTES, extra)

        self.connection.set_context({"zone_id": zone.id})
        response = self.connection.request(url, data=body, method="POST")

        item = response.object["result"]
        record = self._to_record(zone, item)

        return record

    def update_record(self, record, name=None, type=None, data=None, extra=None):
        """
        @inherits: :class:`DNSDriver.update_record`

        Note that for ``extra`` record properties, only the ones specified in
        ``RECORD_UPDATE_ATTRIBUTES`` can be updated. Any non-updatable
        properties are ignored.
        """
        url = "{}/zones/{}/dns_records/{}".format(API_BASE, record.zone.id, record.id)

        content, data = self._normalize_record_data_for_api(type=type, data=data)
        body = {
            "type": record.type if type is None else type,
            "name": record.name if name is None else name,
            "content": content,
            "extra": record.extra or {},
        }

        if data:
            body["data"] = data

        merge_valid_keys(body["extra"], RECORD_UPDATE_ATTRIBUTES, extra)

        self.connection.set_context({"record_id": record.id})
        response = self.connection.request(url, data=body, method="PUT")

        item = response.object["result"]
        record = self._to_record(record.zone, item)

        return record

    def delete_record(self, record):
        url = "{}/zones/{}/dns_records/{}".format(API_BASE, record.zone.id, record.id)

        self.connection.set_context({"record_id": record.id})
        response = self.connection.request(url, method="DELETE")

        item = response.object.get("result", {}).get("id")
        is_deleted = item == record.id

        return is_deleted

    def ex_get_user_account_memberships(self):
        def _ex_get_user_account_memberships(params):
            url = "{}/memberships".format(API_BASE)

            response = self.connection.request(url, params=params)
            return response, response.object["result"]

        return self._paginate(_ex_get_user_account_memberships, self.MEMBERSHIPS_PAGE_SIZE)

    def ex_get_zone_stats(self, zone, interval=30):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_zone_check(self, zones):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_get_ip_threat_score(self, ip):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_get_zone_settings(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_set_zone_security_level(self, zone, level):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_set_zone_cache_level(self, zone, level):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_enable_development_mode(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_disable_development_mode(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_purge_cached_files(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_purge_cached_file(self, zone, url):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_whitelist_ip(self, zone, ip):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_blacklist_ip(self, zone, ip):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_unlist_ip(self, zone, ip):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_enable_ipv6_support(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def ex_disable_ipv6_support(self, zone):
        raise NotImplementedError("not yet implemented in v4 driver")

    def _normalize_record_data_for_api(self, type, data):
        """
        Normalize record data for "special" records such as CAA and SSHFP
        so it can be used with the CloudFlare API.

        Keep ind mind that value for SSHFP record type onluy needs to be
        normalized for the create / update operations.

        On list operation (aka response), actual value is returned
        normally in the "content" attribute.
        """
        cf_data = {}
        if not data:
            return data, cf_data

        if type == RecordType.CAA:
            # Replace whitespace with \t character which CloudFlare API
            # expects
            data = data.replace(" ", "\t")

        elif type == RecordType.SSHFP:
            _fp = data.split(" ")
            cf_data = {"algorithm": _fp[0], "type": _fp[1], "fingerprint": _fp[2]}
            data = None

        return data, cf_data

    def _normalize_record_data_from_api(self, type, data):
        """
        Normalize record data for special records so it's consistent with
        the Libcloud API.
        """
        if not data:
            return data

        if type == RecordType.CAA:
            # CloudFlare uses \t but we normalize it to whitespace so it's
            # consistent across all the drivers.
            data = data.replace("\t", " ")

        return data

    def _to_zone(self, item):
        return Zone(
            id=item["id"],
            domain=item["name"],
            type=CLOUDFLARE_TO_LIBCLOUD_ZONE_TYPE[item["type"]],
            ttl=None,
            driver=self,
            extra={key: item.get(key) for key in ZONE_EXTRA_ATTRIBUTES},
        )

    def _to_record(self, zone, item):
        name = item["name"]
        name = name.replace("." + item["zone_name"], "")
        name = name.replace(item["zone_name"], "")
        name = name or None

        ttl = item.get("ttl")
        if ttl is not None:
            ttl = int(ttl)

        data = self._normalize_record_data_from_api(item["type"], item["content"])

        return Record(
            id=item["id"],
            name=name,
            type=item["type"],
            data=data,
            zone=zone,
            driver=self,
            ttl=ttl,
            extra={key: item.get(key) for key in RECORD_EXTRA_ATTRIBUTES},
        )

    def _paginate(self, get_page, page_size):
        for page in itertools.count(start=1):
            params = {"page": page, "per_page": page_size}

            response, items = get_page(params)

            yield from items

            if self._is_last_page(response):
                break

            if len(items) < page_size:
                break

    def _is_last_page(self, response):
        try:
            result_info = response.object["result_info"]
            last_page = result_info["total_pages"]
            current_page = result_info["page"]
        except KeyError:
            return False

        return current_page == last_page
